Package: MsBackendMgf
Title: Mass Spectrometry Data Backend for Mascot Generic Format (mgf)
        Files
Version: 1.19.0
Authors@R: 
    c(person(given = "RforMassSpectrometry Package Maintainer",
             email = "maintainer@rformassspectrometry.org",
             role = "cre"),
      person(given = "Laurent", family = "Gatto",
             role = "aut",
             email = "laurent.gatto@uclouvain.be",
             comment = c(ORCID = "0000-0002-1520-2268")),
      person(given = "Johannes", family = "Rainer",
             email = "Johannes.Rainer@eurac.edu",
             role = "aut",
             comment = c(ORCID = "0000-0002-6977-7147")),
      person(given = "Sebastian", family = "Gibb",
             comment = c(ORCID = "0000-0001-7406-4443"),
             email = "mail@sebastiangibb.de",
             role = "aut"),
      person(given = "Michael", family = "Witting",
             comment = c(ORCID = "0000-0002-1462-4426"),
             email = "michael.witting@helmholtz-muenchen.de",
             role = "ctb"),
      person(given = "Adriano", family = "Rutz",
             comment = c(ORCID = "0000-0003-0443-9902"),
	     email = "rutz@imsb.biol.ethz.ch",
	     role = "ctb"),
	     person(given = "Corey", family = "Broeckling",
	     comment = c(ORCID = "0000-0002-6158-827X"),
	     email = "Corey.Broeckling@ColoState.edu", role = "ctb"))
Description: Mass spectrometry (MS) data backend supporting import and
	     export of MS/MS spectra data from Mascot Generic Format
	     (mgf) files. Objects defined in this package are supposed to be
	     used with the Spectra Bioconductor package. This package thus
	     adds mgf file support to the Spectra package.
Depends: R (>= 4.0), Spectra (>= 1.5.14)
Imports: ProtGenerics (>= 1.35.3), BiocParallel, S4Vectors, IRanges,
        MsCoreUtils, methods, stats
Suggests: testthat, knitr (>= 1.1.0), roxygen2, BiocStyle (>= 2.5.19),
        rmarkdown
License: Artistic-2.0
Encoding: UTF-8
VignetteBuilder: knitr
BugReports: https://github.com/RforMassSpectrometry/MsBackendMgf/issues
URL: https://github.com/RforMassSpectrometry/MsBackendMgf
biocViews: Infrastructure, Proteomics, MassSpectrometry, Metabolomics,
        DataImport
Roxygen: list(markdown=TRUE)
RoxygenNote: 7.3.2
Collate: 'hidden_aliases.R' 'MsBackendMgf.R' 'functions-mgf.R'
git_url: https://git.bioconductor.org/packages/MsBackendMgf
git_branch: devel
git_last_commit: 3de3a3e
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-10-31
NeedsCompilation: no
Packaged: 2025-10-31 23:00:21 UTC; biocbuild
Author: RforMassSpectrometry Package Maintainer [cre],
  Laurent Gatto [aut] (ORCID: <https://orcid.org/0000-0002-1520-2268>),
  Johannes Rainer [aut] (ORCID: <https://orcid.org/0000-0002-6977-7147>),
  Sebastian Gibb [aut] (ORCID: <https://orcid.org/0000-0001-7406-4443>),
  Michael Witting [ctb] (ORCID: <https://orcid.org/0000-0002-1462-4426>),
  Adriano Rutz [ctb] (ORCID: <https://orcid.org/0000-0003-0443-9902>),
  Corey Broeckling [ctb] (ORCID: <https://orcid.org/0000-0002-6158-827X>)
Maintainer: RforMassSpectrometry Package Maintainer <maintainer@rformassspectrometry.org>

# Generated by roxygen2: do not edit by hand

export(MsBackendAnnotatedMgf)
export(MsBackendMgf)
export(readMgf)
export(readMgfSplit)
exportClasses(MsBackendMgf)
exportMethods(backendInitialize)
exportMethods(export)
exportMethods(spectraVariableMapping)
importClassesFrom(Spectra,MsBackendDataFrame)
importFrom(BiocParallel,SerialParam)
importFrom(BiocParallel,bpmapply)
importFrom(IRanges,NumericList)
importFrom(MsCoreUtils,rbindFill)
importFrom(S4Vectors,DataFrame)
importFrom(Spectra,coreSpectraVariables)
importFrom(methods,as)
importFrom(methods,new)
importFrom(methods,validObject)
importFrom(stats,setNames)
importMethodsFrom(BiocParallel,bplapply)
importMethodsFrom(ProtGenerics,backendInitialize)
importMethodsFrom(ProtGenerics,peaksData)
importMethodsFrom(S4Vectors,metadata)
importMethodsFrom(Spectra,"$")
importMethodsFrom(Spectra,"$<-")
importMethodsFrom(Spectra,"spectraData<-")
importMethodsFrom(Spectra,backendInitialize)
importMethodsFrom(Spectra,export)
importMethodsFrom(Spectra,spectraData)
importMethodsFrom(Spectra,spectraNames)
importMethodsFrom(Spectra,spectraVariableMapping)
importMethodsFrom(Spectra,spectraVariables)

# Mass Spectrometry Data Backend for Mascot Generic Format (mgf) Files

[![Project Status: Active – The project has reached a stable, usable state and is being actively developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![R-CMD-check-bioc](https://github.com/RforMassSpectrometry/MsBackendMgf/workflows/R-CMD-check-bioc/badge.svg)](https://github.com/RforMassSpectrometry/MsBackendMgf/actions?query=workflow%3AR-CMD-check-bioc)
[![codecov](https://codecov.io/gh/rformassspectrometry/MsBackendMgf/branch/main/graph/badge.svg?token=R1VB6BmWAK)](https://codecov.io/gh/rformassspectrometry/MsBackendMgf)
[![license](https://img.shields.io/badge/license-Artistic--2.0-brightgreen.svg)](https://opensource.org/licenses/Artistic-2.0)
[![years in bioc](http://bioconductor.org/shields/years-in-bioc/MsBackendMgf.svg)](https://bioconductor.org/packages/release/bioc/html/MsBackendMgf.html)
[![Ranking by downloads](http://bioconductor.org/shields/downloads/release/MsBackendMgf.svg)](https://bioconductor.org/packages/stats/bioc/MsBackendMgf/)
[![build release](http://bioconductor.org/shields/build/release/bioc/MsBackendMgf.svg)](https://bioconductor.org/checkResults/release/bioc-LATEST/MsBackendMgf/)
[![build devel](http://bioconductor.org/shields/build/devel/bioc/MsBackendMgf.svg)](https://bioconductor.org/checkResults/devel/bioc-LATEST/MsBackendMgf/)

The `MsBackendMgf` package provides functionality to import and handle
MS/MS spectrum data from Mascot Generic Format
([mgf](http://www.matrixscience.com/help/data_file_help.html)) files.
The package defines the `MsBackendMgf` backend which can be used to
import and use MS2 spectrum data from mgf files with the
[Spectra](https://github.com/rformassspectrometry/Spectra) R package.

For more information see the package
[homepage](https://rformassspectrometry.github.io/MsBackendMgf).


# Installation

The package can be installed with

```r
install.packages("BiocManager")
BiocManager::install("MsBackendMgf")
```


# Contributions

Contributions are highly welcome and should follow the [contribution
guidelines](https://rformassspectrometry.github.io/RforMassSpectrometry/articles/RforMassSpectrometry.html#contributions).
Also, please check the coding style guidelines in the [RforMassSpectrometry
vignette](https://rformassspectrometry.github.io/RforMassSpectrometry/articles/RforMassSpectrometry.html).

library(testthat)
library(MsBackendMgf)

fls <- dir(system.file("extdata", package = "MsBackendMgf"),
           full.names = TRUE, pattern = "mgf$")[1:2]

test_check("MsBackendMgf")

## Run tests defined in test suites from the Spectra package.
fls <- dir(system.file("extdata", package = "MsBackendMgf"),
           full.names = TRUE, pattern = "mgf$")[1:2]
be <- MsBackendMgf()
be <- backendInitialize(be, fls)

## Run the MsBackend spectra variable test suite
test_suite <- system.file("test_backends", "test_MsBackend",
                          package = "Spectra")

## Run single test file.
res <- test_dir(test_suite, stop_on_failure = TRUE)

#' @title Internal page for hidden aliases
#'
#' @aliases [,MsBackendDataFrame-method
#'
#' @description
#'
#' For S4 methods that require a documentation entry but only clutter the index.
#'
#' @usage NULL
#' @format NULL
#' @keywords internal
#' @docType methods
hidden_aliases <- NULL

#' @include hidden_aliases.R
NULL

#' @title MS data backend for MGF files
#'
#' @aliases MsBackendMgf-class
#'
#' @description
#'
#' The `MsBackendMgf` class supports import and export of MS/MS spectra data
#' from/to files in Mascot Generic Format
#' ([mgf](http://www.matrixscience.com/help/data_file_help.html))
#' files. After initial import, the full MS data is kept in
#' memory. `MsBackendMgf` extends the [Spectra::MsBackendDataFrame()] backend
#' directly and supports thus the [Spectra::applyProcessing()] function to make
#' data manipulations persistent.
#'
#' The `MsBackendAnnotatedMgf` class supports import of data from MGF files
#' that provide, in addition to the *m/z* and intensity values, also
#' additional annotations/metadata for each mass peak. For such MGF files it
#' is expected that each line contains information from a single mass peak,
#' separated by a white space (blank). The first two elements are expected to
#' be the peak's *m/z* and intensity values, while each additional element is
#' considered an annotation for this specific peak. See examples below for the
#' format of a supported MGF file. The `backendInitialize()` method of
#' `MsBackendAnnotatedMgf` does not support parameter `nlines`. Also, import
#' of data can be considerably slower compared to the standard `MsBackendMgf`
#' backend, because of the additionally required parsing of peak annotations.
#' Peaks information in MGF files are not named, thus, additional peaks
#' annotations are named using the standard naming convention for column named
#' of data frames: the first peaks annotation is called `"V1"`, the second (if
#' available) `"V2"` and so on.
#'
#' New objects are created with the `MsBackendMgf()` or
#' `MsBackendAnnotatedMgf()` function. The `backendInitialize()` method has to
#' be subsequently called to initialize the object and import the MS/MS data
#' from (one or more) MGF files.
#'
#' The `MsBackendMgf` backend provides an `export` method that allows to export
#' the data from the `Spectra` object (parameter `x`) to a file in mgf format.
#' See the package vignette for details and examples.
#'
#' Default mappings from fields in the MGF file to spectra variable names are
#' provided by the `spectraVariableMapping()` function. This function returns a
#' named character vector were names are the spectra variable names and the
#' values the respective field names in the MGF files. This named character
#' vector is submitted to the import and export function with parameter
#' `mapping`. It is also possible to pass own mappings (e.g. for special
#' MGF dialects) with the `mapping` parameter.
#'
#' @param object Instance of `MsBackendMgf` class.
#'
#' @param file `character(1)` with the (full) file name to which the data
#'     should be exported.
#'
#' @param files `character` with the (full) file name(s) of the mgf file(s)
#'     from which MS/MS data should be imported.
#'
#' @param format for `spectraVariableMapping()`: `character(1)` defining the
#'     format to be used. Currently only `format = "mgf"` is supported.
#'
#' @param mapping for `backendInitialize()` and `export`: named `character`
#'     vector allowing to specify how fields from the MGF file should be
#'     renamed. Names are supposed to be the spectra variable name and
#'     values of the vector the field names in the MGF file. See output of
#'     `spectraVariableMapping()` for the expected format and examples
#'     below or description above for details.
#'
#' @param nlines for `backendInitialize()` of `MsBackendMgf`: `integer(1)`
#'     defining the number of lines that should be imported and processed
#'     from the MGF file(s).
#'     By default (`nlines = -1L`) the full file is imported and processed at
#'     once. If set to a positive integer, the data is imported and processed
#'     *chunk-wise* using [readMgfSplit()].
#'
#' @param exportTitle `logical(1)` whether the *TITLE* field should be included
#'     in the exported MGF file. If `TRUE` (the default) a `spectraVariable`
#'     called `"TITLE"` will be used, if no such variable is present either the
#'     `spectraNames(object)` will be used or, if they are empty, a title will
#'     be generated including the MS level, retention time and acquisition
#'     number of the spectrum.
#'
#' @param x for `export()`: an instance of [Spectra::Spectra()] class with the
#'     data that should be exported.
#'
#' @param BPPARAM Parameter object defining the parallel processing
#'     setup. If parallel processing is enabled (with `BPPARAM` different than
#'     `SerialParam()`, the default) and length of `files` is larger than one,
#'     import is performed in parallel on a per-file basis. If data is to be
#'     imported from a single file (i.e., length of `files` is one), parsing
#'     of the imported file is performed in parallel. See also
#'     [BiocParallel::SerialParam()] for information on available parallel
#'     processing setup options.
#'
#' @param ... Currently ignored.
#'
#' @author Laurent Gatto, Corey Broeckling and Johannes Rainer
#'
#' @importClassesFrom Spectra MsBackendDataFrame
#'
#' @exportClass MsBackendMgf
#'
#' @name MsBackendMgf
#'
#' @return
#'
#' See description above.
#'
#' @examples
#'
#' library(BiocParallel)
#' #' Getting the file names of all example MGF files from MsBackendMgf
#' fls <- dir(system.file("extdata", package = "MsBackendMgf"),
#'     full.names = TRUE, pattern = "^spectra(.*).mgf$")
#'
#' ## Create an MsBackendMgf backend and import data from test mgf files.
#' be <- backendInitialize(MsBackendMgf(), fls)
#' be
#'
#' be$msLevel
#' be$intensity
#' be$mz
#'
#' ## The spectra variables that are available; note that not all of them
#' ## have been imported from the MGF files.
#' spectraVariables(be)
#'
#' ## The variable "TITLE" represents the title of the spectrum defined in the
#' ## MGF file
#' be$TITLE
#'
#' ## The default mapping of MGF fields to spectra variables is provided by
#' ## the spectraVariableMapping function
#' spectraVariableMapping(MsBackendMgf())
#'
#' ## We can provide our own mapping e.g. to map the MGF field "TITLE" to a
#' ## variable named "spectrumName":
#' map <- c(spectrumName = "TITLE", spectraVariableMapping(MsBackendMgf()))
#' map
#'
#' ## We can then pass this mapping with parameter `mapping` to the
#' ## backendInitialize method:
#' be <- backendInitialize(MsBackendMgf(), fls, mapping = map)
#'
#' ## The title is now available as variable named spectrumName
#' be$spectrumName
#'
#' ## Next we create a Spectra object with this data
#' sps <- Spectra(be)
#'
#' ## We can use the 'MsBackendMgf' also to export spectra data in mgf format.
#' out_file <- tempfile()
#' export(sps, backend = MsBackendMgf(), file = out_file, map = map)
#'
#' ## The first 20 lines of the generated file:
#' readLines(out_file, n = 20)
#'
#' ## Next we add a new spectra variable to each spectrum
#' sps$spectrum_idx <- seq_along(sps)
#'
#' ## This new spectra variable will also be exported to the mgf file:
#' export(sps, backend = MsBackendMgf(), file = out_file, map = map)
#' readLines(out_file, n = 20)
#'
#' ####
#' ## Annotated MGF
#'
#' ## An example of a supported annotated MGF file
#' fl <- system.file("extdata", "xfiora.mgf", package = "MsBackendMgf")
#'
#' ## Lines with peak data start with a numeric and information is
#' ## separated by a whitespace. The first two elements are the peak's m/z
#' ## and intensity while any additional information is considered as
#' ## annotation. Information for each peak is provided in one line.
#' readLines(fl)
#'
#' ## Importing the data using an `MsBackendAnnotatedMgf`
#' ba <- backendInitialize(MsBackendAnnotatedMgf(), fl)
#' ba
#'
#' ## An additional peaks variable is available.
#' peaksVariables(ba)
#'
#' ba$V1
#'
#' ## The length of such peaks variables is the same as the length of the
#' ## m/z or intensity values, i.e. each peak has one value (with the value
#' ## being `NA` if missing).
#' length(ba$V1[[1L]])
#' length(ba$mz[[1L]])
#'
#' ## Extracting the peaks data from a `Spectra` with a `MsBackendAnnotatedMgf`
#' s <- Spectra(ba)
#' pd <- peaksData(s, peaksVariables(ba))[[1L]]
#' head(pd)
#' class(pd)
NULL

setClass("MsBackendMgf",
         contains = "MsBackendDataFrame",
         prototype = prototype(spectraData = DataFrame(),
                               readonly = FALSE,
                               version = "0.1"))

#' @importMethodsFrom Spectra spectraData<- $<- $
#'
#' @importFrom BiocParallel SerialParam
#'
#' @importMethodsFrom ProtGenerics backendInitialize
#'
#' @importMethodsFrom Spectra backendInitialize
#'
#' @importMethodsFrom BiocParallel bplapply
#'
#' @importFrom methods validObject
#'
#' @exportMethod backendInitialize
#'
#' @rdname MsBackendMgf
setMethod("backendInitialize", signature = "MsBackendMgf",
          function(object, files, mapping = spectraVariableMapping(object),
                   nlines = -1L, ..., BPPARAM = SerialParam()) {
              if (missing(files) || !length(files))
                  stop("Parameter 'files' is mandatory for ", class(object))
              if (!is.character(files))
                  stop("Parameter 'files' is expected to be a character vector",
                       " with the files names from where data should be",
                       " imported")
              if (!is.numeric(nlines))
                  stop("'nlines' needs to be an integer")
              nlines <- as.integer(nlines)
              files <- normalizePath(files)
              if (any(!file.exists(files)))
                  stop("file(s) ",
                       paste(files[!file.exists(files)], collapse = ", "),
                       " not found")
              ## Import data and rbind.
              message("Start data import from ", length(files), " files ... ",
                      appendLF = FALSE)
              if (nlines > 0)
                  FUN <- readMgfSplit
              else FUN <- readMgf
              if (length(files) > 1) {
                  res <- bplapply(files, FUN = FUN, mapping = mapping,
                                  nlines = nlines, BPPARAM = BPPARAM)
                  res <- do.call(rbindFill, res)
              } else
                  res <- FUN(files, mapping = mapping, nlines = nlines,
                             BPPARAM = BPPARAM)
              message("done")
              res@metadata <- list()
              spectraData(object) <- res
              object$dataStorage <- "<memory>"
              object$centroided <- TRUE
              validObject(object)
              object
          })

#' @rdname MsBackendMgf
#'
#' @importFrom methods new
#'
#' @export MsBackendMgf
MsBackendMgf <- function() {
    new("MsBackendMgf")
}

#' @importMethodsFrom Spectra spectraVariableMapping
#'
#' @exportMethod spectraVariableMapping
#'
#' @rdname MsBackendMgf
setMethod("spectraVariableMapping", "MsBackendMgf",
          function(object, format = c("mgf")) {
              switch(match.arg(format),
                     "mgf" = c(
                         rtime = "RTINSECONDS",
                         acquisitionNum = "SCANS",
                         precursorMz = "PEPMASS",
                         precursorIntensity = "PEPMASSINT",
                         precursorCharge = "CHARGE"
                     )
                     )
          })

#' @importMethodsFrom Spectra export
#'
#' @exportMethod export
#'
#' @rdname MsBackendMgf
setMethod("export", "MsBackendMgf",
          function(object, x, file = tempfile(),
                   mapping = spectraVariableMapping(object),
                   exportTitle = TRUE, ...) {
              if (missing(x))
                  stop("Required parameter 'x' is missing. 'x' should be a ",
                       "'Spectra' object with the full spectra data.")
              if (!inherits(x, "Spectra"))
                  stop("Parameter 'x' is supposed to be a 'Spectra' object ",
                       "with the full spectra data to be exported.")
              .export_mgf(x = x, con = file, mapping = mapping,
                          exportTitle = exportTitle)
          })

setClass("MsBackendAnnotatedMgf",
         contains = "MsBackendMgf",
         prototype = prototype(spectraData = DataFrame(),
                               readonly = FALSE,
                               version = "0.1"))

#' @rdname MsBackendMgf
#'
#' @importMethodsFrom S4Vectors metadata
setMethod("backendInitialize", signature = "MsBackendAnnotatedMgf",
          function(object, files, mapping = spectraVariableMapping(object),
                   ..., BPPARAM = SerialParam()) {
              if (missing(files) || !length(files))
                  stop("Parameter 'files' is mandatory for ", class(object))
              if (!is.character(files))
                  stop("Parameter 'files' is expected to be a character vector",
                       " with the files names from where data should be",
                       " imported")
              files <- normalizePath(files)
              if (any(!file.exists(files)))
                  stop("file(s) ",
                       paste(files[!file.exists(files)], collapse = ", "),
                       " not found")
              ## Import data and rbind.
              message("Start data import from ", length(files), " files ... ",
                      appendLF = FALSE)
              if (length(files) > 1) {
                  res <- bplapply(files, FUN = readMgf, mapping = mapping,
                                  annotated = TRUE, BPPARAM = BPPARAM)
                  pcol <- unique(unlist(lapply(res, metadata)))
                  res <- do.call(rbindFill, res)
              } else {
                  res <- readMgf(files, mapping = mapping, annotated = TRUE,
                                 BPPARAM = BPPARAM)
                  pcol <- unique(unlist(metadata(res)))
              }
              message("done")
              res@metadata <- list()
              spectraData(object) <- res
              object@peaksVariables <- c("mz", "intensity", pcol)
              object$dataStorage <- "<memory>"
              object$centroided <- TRUE
              validObject(object)
              object
          })

#' @rdname MsBackendMgf
#'
#' @export MsBackendAnnotatedMgf
MsBackendAnnotatedMgf <- function() {
    new("MsBackendAnnotatedMgf")
}

test_that("backendInitialize,MsBackendMgf works", {
    fls <- dir(system.file("extdata", package = "MsBackendMgf"),
               full.names = TRUE, pattern = "mgf$")[1:2]
    be <- MsBackendMgf()

    ## Import a single file.
    res1 <- backendInitialize(be, fls[1])
    n1 <- length(res1) ## 3
    expect_identical(length(res1), n1)
    expect_identical(res1$dataStorage, rep("<memory>", n1))
    expect_identical(res1$dataOrigin, rep(normalizePath(fls[1]), n1))
    expect_identical(res1$msLevel, rep(2L, n1))
    expect_identical(lengths(res1$mz), c(14L, 21L, 14L))

    res1_b <- backendInitialize(be, fls[1], nlines = 30)
    expect_equal(res1, res1_b)

    res2 <- backendInitialize(be, fls[2])
    n2 <- length(res2) ## 4
    expect_equal(n2, 4)

    ## Import multiple files.
    res_all <- backendInitialize(be, fls)
    expect_true(length(res_all) == n1 + n2)
    expect_identical(res_all[1]$mz, res1[1]$mz)
    expect_identical(res_all[n1 + 1]$mz, res2[1]$mz)
    expect_true(all(res_all$msLevel == 2L))
    expect_identical(res_all$dataOrigin,
                     c(rep(normalizePath(fls[1]), n1),
                       rep(normalizePath(fls[2]), n2)))
    expect_true(is.integer(res_all@spectraData$msLevel))

    ## errors
    expect_error(backendInitialize(be), "'files' is mandatory")
    expect_error(backendInitialize(be, 4), "expected to be a character")
    expect_error(backendInitialize(be, "a"), "a not found")
    expect_error(backendInitialize(be, fls[1], nlines = "a"), "integer")
    expect_error(backendInitialize(be, "a"), "not found")
})

test_that("spectraVariableMapping works", {
    res <- spectraVariableMapping(MsBackendMgf())
    expect_true(is.character(res))
    expect_true(length(names(res)) == length(res))

    expect_error(spectraVariableMapping(MsBackendMgf(), format = "other"))
})

test_that("mixed MS level import works", {

  fls <- dir(system.file("extdata", package = "MsBackendMgf"),
             full.names = TRUE, pattern = "mgf$")[4]

  custom_mapping <- c(rtime = "RTINSECONDS",
                      acquisitionNum = "SCANS",
                      precursorMz = "PEPMASS",
                      precursorIntensity = "PEPMASSINT",
                      precursorCharge = "CHARGE",
                      msLevel = "MSLEVEL")

  res <- Spectra(fls,
                 source = MsBackendMgf(),
                 backend = MsBackendDataFrame(),
                 mapping = custom_mapping)

  expect_identical(length(res), 2L)
  expect_identical(res$msLevel, c(1L, 2L))
})

test_that("export,MsBackendMgf works", {
    spd <- DataFrame(msLevel = c(2L, 2L, 2L), rtime = c(1, 2, 3))
    spd$mz <- list(c(12, 14, 45, 56), c(14.1, 34, 56.1), c(12.1, 14.15, 34.1))
    spd$intensity <- list(c(10, 20, 30, 40), c(11, 21, 31), c(12, 22, 32))

    sps <- Spectra(spd)

    fl <- tempfile()
    export(MsBackendMgf(), sps, file = fl)
    res <- backendInitialize(MsBackendMgf(), fl)
    expect_equal(rtime(res), rtime(sps))
    expect_equal(peaksData(res), peaksData(sps@backend))

    sps$precursorCharge <- c(2L, NA_integer_, -4L)
    export(MsBackendMgf(), sps, file = fl)
    res <- backendInitialize(MsBackendMgf(), fl)
    expect_equal(precursorCharge(res), precursorCharge(sps))

    export(MsBackendMgf(), sps, file = fl, exportTitle = FALSE)
    res <- readLines(fl)
    expect_true(length(grep("TITLE", res)) == 0)

    sps$TITLE <- c("a", "b", "c")
    export(MsBackendMgf(), sps, file = fl, exportTitle = FALSE)
    res <- readLines(fl)
    expect_true(length(grep("TITLE", res)) == 0)

    spectraNames(sps) <- c("d", "e", "f")
    export(MsBackendMgf(), sps, file = fl, exportTitle = FALSE)
    res <- readLines(fl)
    expect_true(length(grep("TITLE", res)) == 0)

    spectraNames(sps) <- NULL
    sps$TITLE <- NULL
    export(MsBackendMgf(), sps, file = fl, exportTitle = TRUE)
    res <- readLines(fl)
    expect_true(length(grep("TITLE=msLevel", res)) > 0)

    spectraNames(sps) <- paste0("yes", seq_along(sps))
    sps$TITLE <- NULL
    export(MsBackendMgf(), sps, file = fl, exportTitle = TRUE)
    res <- readLines(fl)
    expect_true(length(grep("TITLE=yes", res)) > 0)

    expect_error(export(MsBackendMgf(), file = fl), "missing")
    expect_error(export(MsBackendMgf(), x = spd, file = fl), "spectra data to")

    sps$lst <- list(1:3, c(4, 5, 2), c("a", "b"))
    expect_error(export(MsBackendMgf(), sps, file = fl), "multiple elements")

    sps$fail <- IRanges::NumericList(sps$mz, compress = FALSE)
    expect_error(export(MsBackendMgf(), sps, file = fl), "multiple elements")
})

test_that("Spectra works with MsBackendMgf", {
    s <- Spectra(fls, source = MsBackendMgf())
    be <- backendInitialize(MsBackendMgf(), fls)
    expect_equal(be$mz, s$mz)
})

test_that("MsBackendAnnotatedMgf works", {
    expect_s4_class(MsBackendAnnotatedMgf(), "MsBackendAnnotatedMgf")

    fls <- dir(system.file("extdata", package = "MsBackendMgf"),
               full.names = TRUE, pattern = "mgf$")
    b <- grep("fiora", fls, value = TRUE)
    be <- backendInitialize(MsBackendAnnotatedMgf(), c(b, b))
    expect_s4_class(be, "MsBackendAnnotatedMgf")
    expect_true(length(be) == 2L)
    expect_equal(be[1]$mz, be[2]$mz)
    expect_equal(peaksVariables(be), c("mz", "intensity", "V1"))
    expect_equal(be[1]$V1, be[2]$V1)
    pks <- peaksData(be, c("mz", "intensity", "V1"))
    expect_true(is.data.frame(pks[[1L]]))
    expect_equal(colnames(pks[[1L]]), c("mz", "intensity", "V1"))

    be2 <- backendInitialize(MsBackendAnnotatedMgf(), b)
    expect_equal(be$V1[[1L]], be2$V1[[1L]])
    expect_s4_class(be2, "MsBackendAnnotatedMgf")
    expect_equal(peaksVariables(be2), c("mz", "intensity", "V1"))

    ## errors
    expect_error(backendInitialize(be), "'files' is mandatory")
    expect_error(backendInitialize(be, 4), "expected to be a character")
    expect_error(backendInitialize(be, "a"), "a not found")
})

##' @title Reading MGF files
##'
##' @description
##'
##' The `readMgf()` function imports the data from a file in MGF format reading
##' all specified fields and returning the data as a [S4Vectors::DataFrame()].
##'
##' For very large MGF files the `readMgfSplit()` function might be used
##' instead. In contrast to the `readMgf()` functions, `readMgfSplit()` reads
##' only `nlines` lines from an MGF file at once reducing thus the memory
##' demand (at the cost of a lower performance, compared to `readMgf()`).
##'
##' @param f `character(1)` with the path to an mgf file.
##'
##' @param msLevel `numeric(1)` with the MS level. Default is 2.
##'
##' @param mapping named `character` vector to rename mgf fields to spectra
##'     variables.
##'
##' @param annotated For `readMgf()`: `logical(1)` whether the MGF file
##'     contains additional peak annotations. See examples below or the
##'     documentation for [MsBackendAnnotatedMgf()] for information on the
##'     expected format.
##'
##' @param nlines for `readMgfSplit()`: `integer(1)` with the number of lines
##'     that should be imported and parsed in each iteration.
##'
##' @param BPPARAM parallel processing setup that should be used. Only the
##'     parsing of the imported MGF file is performed in parallel.
##'
##' @param ... Additional parameters, currently ignored.
##'
##' @return
##'
##' A `DataFrame` with each row containing the data from one spectrum
##' in the MGF file. m/z and intensity values are available in columns `"mz"`
##' and `"intensity"` in a list representation. For `readMgf()` with
##' `annotated = TRUE` also all peaks annotation columns (named `"V1", etc)
##' are provided in a list representation, with the lengths of elements
##' matching those of `"mz"` or `"intensity"`.
##'
##' @export
##'
##' @importFrom Spectra coreSpectraVariables
##'
##' @importFrom S4Vectors DataFrame
##'
##' @importFrom IRanges NumericList
##'
##' @importFrom MsCoreUtils rbindFill
##'
##' @importFrom methods as
##'
##' @importFrom BiocParallel SerialParam bpmapply
##'
##' @author Laurent Gatto, Johannes Rainer, Sebastian Gibb, Corey Broeckling
##'
##' @examples
##'
##' fls <- dir(system.file("extdata", package = "MsBackendMgf"),
##'     full.names = TRUE, pattern = "mgf$")[1L]
##'
##' readMgf(fls)
##'
##' ## Annotated MGF
##' fl <- system.file("extdata", "xfiora.mgf", package = "MsBackendMgf")
##' res <- readMgf(fl, annotated = TRUE)
##' colnames(res)
##' res$V1
readMgf <- function(f, msLevel = 2L,
                    mapping = spectraVariableMapping(MsBackendMgf()),
                    annotated = FALSE, ...,
                    BPPARAM = SerialParam()) {
    requireNamespace("MsBackendMgf", quietly = TRUE)
    if (length(f) != 1L)
        stop("Please provide a single mgf file.")
    ## Note: using readLines instead has some performance advantages
    ## (few seconds) for very large files
    mgf <- scan(file = f, what = "",
                sep = "\n", quote = "",
                allowEscapes = FALSE,
                quiet = TRUE)

    ## From http://www.matrixscience.com/help/data_file_help.html#GEN
    ## Comment lines beginning with one of the symbols #;!/ can be
    ## included, but only outside of the BEGIN IONS and END IONS
    ## statements that delimit an MS/MS dataset.

    begin <- grep("BEGIN IONS", mgf, fixed = TRUE) + 1L
    end <- grep("END IONS", mgf, fixed = TRUE) - 1L

    if (annotated) fun_extract_mgf <- .extract_mgf_spectrum_with_annotations
    else fun_extract_mgf <- .extract_mgf_spectrum

    res <- bpmapply(begin, end, FUN = function(b, e, mgf)
        fun_extract_mgf(mgf[b:e]), MoreArgs = list(mgf = mgf),
        SIMPLIFY = FALSE, USE.NAMES = FALSE, BPPARAM = BPPARAM)
    res <- rbindFill(res)
    if (annotated) {
        p <- as.factor(rep(seq_len(nrow(res)), lengths(res$mz)))
        anns <- rbindFill(res$ann_mat)
        res$ann_mat <- NULL
        pcol <- colnames(anns)
        for (a in pcol)
            res <- do.call(
                "$<-", list(res, name = a, value = unname(split(anns[[a]], p))))
    } else pcol <- character()

    if ("CHARGE" %in% colnames(res))
        res$CHARGE <- .format_charge(res$CHARGE)

    idx <- match(colnames(res), mapping)
    not_na <- !is.na(idx)
    if (any(not_na))
        colnames(res)[not_na] <- names(mapping)[idx][not_na]

    spv <- coreSpectraVariables()
    spv <- spv[!names(spv) %in% c("mz", "intensity")]
    for (i in seq_along(res)) {
        if (all(lengths(res[[i]]) == 1))
            res[[i]] <- unlist(res[[i]])
        if (any(col <- names(spv) == colnames(res)[i]))
            res[[i]] <- as(res[[i]], spv[col][1])
    }

    res <- as(res, "DataFrame")
    if (annotated)
        res@metadata <- list(pcol)
    res$mz <- IRanges::NumericList(res$mz, compress = FALSE)
    res$intensity <- IRanges::NumericList(res$intensity, compress = FALSE)
    res$dataOrigin <- f
    if(!"msLevel" %in% colnames(res))
      res$msLevel <- as.integer(msLevel)

    res
}

#' @export
#'
#' @rdname readMgf
readMgfSplit <- function(f, msLevel = 2L,
                         mapping = spectraVariableMapping(MsBackendMgf()),
                         nlines = 100000, BPPARAM = SerialParam(), ...) {
    requireNamespace("MsBackendMgf", quietly = TRUE)
    if (length(f) != 1L)
        stop("Please provide a single mgf file.")
    if (nlines < 1)
        stop("'nlines' has to be an integer > 1.")
    keep_reading <- TRUE
    nskip <- 0L
    sp <- list()
    while (keep_reading) {
        mgf <- scan(file = f, what = "",
                    sep = "\n", quote = "",
                    allowEscapes = FALSE,
                    skip = nskip,
                    nlines = nlines,
                    quiet = TRUE,
                    blank.lines.skip = FALSE)
        if (length(mgf) < nlines)
            keep_reading <- FALSE

        begin <- grep("BEGIN IONS", mgf, fixed = TRUE) + 1L
        end <- grep("END IONS", mgf, fixed = TRUE) - 1L
        if (!length(end))
            stop("Could not find 'END IONS'. ",
                 "Consider increasing the value for 'nlines'")
        nskip <- nskip + end[length(end)] + 1L
        begin <- begin[seq_along(end)]
        ## This append operation is very costly; need an alternative.
        sp <- c(sp,
                bplapply(
                    seq_along(begin), function(i, mgf, begin, end) {
                        .extract_mgf_spectrum(mgf[begin[i]:end[i]])},
                    mgf = mgf, begin = begin, end = end,
                    BPPARAM = BPPARAM)
                )
    }
    res <- rbindFill(sp)

    if ("CHARGE" %in% colnames(res))
        res$CHARGE <- .format_charge(res$CHARGE)

    idx <- match(colnames(res), mapping)
    not_na <- !is.na(idx)
    if (any(not_na))
        colnames(res)[not_na] <- names(mapping)[idx][not_na]

    spv <- coreSpectraVariables()
    spv <- spv[!names(spv) %in% c("mz", "intensity")]
    for (i in seq_along(res)) {
        if (all(lengths(res[[i]]) == 1))
            res[[i]] <- unlist(res[[i]])
        if (any(col <- names(spv) == colnames(res)[i]))
            res[[i]] <- as(res[[i]], spv[col][1])
    }

    res <- as(res, "DataFrame")
    res$mz <- IRanges::NumericList(res$mz, compress = FALSE)
    res$intensity <- IRanges::NumericList(res$intensity, compress = FALSE)
    res$dataOrigin <- f
    if(!"msLevel" %in% colnames(res)) {
      res$msLevel <- as.integer(msLevel)
    }

    res
}

##' @description
##'
##' Extract **all** fields from the MGF eventually renaming the field names to
##' the spectra variable names specified with `mapping`.
##'
##' @param mgf `character()` of lines defining a spectrum in mgf
##'     format.
##'
##' @author Laurent Gatto, Johannes Rainer
##'
##' @importFrom stats setNames
##'
##' @noRd
.extract_mgf_spectrum <- function(mgf) {
    ## grep description
    desc.idx <- grep("=", mgf, fixed = TRUE)
    desc <- mgf[desc.idx]

    spec <- strsplit(mgf[-desc.idx], "[[:space:]]+", perl = TRUE)
    if (!length(spec) || length(spec[[1L]]) == 1L)
        ms <- matrix(numeric(), ncol = 2L)
    else
        ms <- matrix(
            as.double(unlist(spec, use.names = FALSE, recursive = FALSE)),
            ncol = length(spec[[1L]]), byrow = TRUE)

    if(.is_unsorted(ms))
        ms <- ms[order(ms[, 1L]), , drop = FALSE]

    r <- regexpr("=", desc, fixed = TRUE)
    desc <- setNames(substring(desc, r + 1L, nchar(desc)),
                     substring(desc, 1L, r - 1L))

    desc[c("PEPMASS", "PEPMASSINT")] <-
        strsplit(desc["PEPMASS"], "[[:space:]]+", perl = TRUE)[[1L]][c(1L, 2L)]

    res <- as.data.frame.matrix(matrix(desc, nrow = 1,
                                       dimnames = list(NULL, names(desc))))
    res$mz <- list(ms[, 1L])
    res$intensity <- list(ms[, 2L])
    res
}

#' @title Process MGF files with peak annotations
#'
#' @description
#'
#' Import MS data from MGF files that provide also annotations for individual
#' mass peaks.
#'
#' Expected format of the peak information:
#'
#' - lines with peak information are expected to start with a number (no white
#'   space)
#' - each line is expected to represent one mass peak
#' - the first two elements per line are expected to be the m/z and intensity
#'   values
#' - all following elements represent peak annotation/metadata. For multiple
#'   spectra, they are assumed to be in the same order, i.e., if 3 annotations
#'   are provided for a peak, they are expected to be in the order
#'   annotation_1<white space>annotation_2<white space>annotation_3 for all
#'   spectra. And it is expected that all are provided.
#' - annotations are interpreted as character strings.
#'
#' @author Johannes Rainer and Corey Broeckling
#'
#' @noRd
.extract_mgf_spectrum_with_annotations <- function(mgf) {
    mz <- numeric()
    int <- numeric()
    ann <- as.data.frame(matrix(NA_character_, ncol = 0, nrow = 0))
    ## find peaks
    pks_idx <- grep("^[0-9]", mgf)
    l <- length(pks_idx)
    if (l) {
        pks <- strsplit(mgf[pks_idx], "[[:space:]]+", perl = TRUE)
        ls <- lengths(pks)
        ml <- max(ls)
        if (ml > 2) {
            mli <- 3:ml
            ann <- as.data.frame(
                do.call(rbind, lapply(pks, function(z) z[mli])))
        } else
            ann <- as.data.frame(matrix(NA_character_, ncol = 0, nrow = l))
        mz <- as.numeric(vapply(pks, `[`, NA_character_, 1L, USE.NAMES = FALSE))
        int <-as.numeric(vapply(pks, `[`, NA_character_, 2L, USE.NAMES = FALSE))
        if (is.unsorted(mz)) {
            idx <- order(mz)
            mz <- mz[idx]
            int <- int[idx]
        }
        mgf <- mgf[-pks_idx]
    }
    ## grep description
    desc <- grep("=", mgf, fixed = TRUE, value = TRUE)

    r <- regexpr("=", desc, fixed = TRUE)
    desc <- setNames(substring(desc, r + 1L, nchar(desc)),
                     substring(desc, 1L, r - 1L))

    desc[c("PEPMASS", "PEPMASSINT")] <-
        strsplit(desc["PEPMASS"], "[[:space:]]+", perl = TRUE)[[1L]][c(1L, 2L)]

    res <- as.data.frame.matrix(matrix(desc, nrow = 1,
                                       dimnames = list(NULL, names(desc))))
    res$mz <- list(mz)
    res$intensity <- list(int)
    res$ann_mat <- list(ann)
    res
}

.is_unsorted <- function(x) {
    nrow(x) && is.unsorted(x[, 1L])
}

#' Format MGF charge string into an integer compatible format.
#'
#' @param x `character`
#' @return `character`, charge without +/- at the end but - as prefix if needed
#' @noRd
.format_charge <- function(x) {
    res <- sub("[+-]", "", x, perl = TRUE)
    negs <- which(endsWith(x, "-"))
    if (length(negs))
        res[negs] <- paste0("-", res[negs])
    res
}

#' @description
#'
#' Function to export a `Spectra` object in MGF format to `con`.
#'
#' @param x `Spectra`
#'
#' @param con output file.
#'
#' @param mapping named `character` vector that maps from `spectraVariables`
#'    (i.e. `names(mapping)`) to the variable name that should be used in the
#'    MGF file.
#'
#' @author Johannes Rainer
#'
#' @importMethodsFrom Spectra spectraVariables spectraNames spectraData
#'
#' @importMethodsFrom ProtGenerics peaksData
#'
#' @noRd
#'
#' @examples
#'
#' spd <- DataFrame(msLevel = c(2L, 2L, 2L), rtime = c(1, 2, 3))
#' spd$mz <- list(c(12, 14, 45, 56), c(14.1, 34, 56.1), c(12.1, 14.15, 34.1))
#' spd$intensity <- list(c(10, 20, 30, 40), c(11, 21, 31), c(12, 22, 32))
#'
#' sps <- Spectra(spd)
#'
#' .export_mgf(sps)
.export_mgf <- function(x, con = stdout(),
                        mapping = spectraVariableMapping(MsBackendMgf()),
                        exportTitle = TRUE) {
    spv <- spectraVariables(x)
    spd <- spectraData(x, spv[!(spv %in% c("dataOrigin", "dataStorage"))])
    col_not_ok <- !vapply(spd, function(z) is.vector(z) & !is.list(z),
                          logical(1))
    if (any(col_not_ok))
        stop("Column(s) ", paste(colnames(spd)[col_not_ok], collapse = ", "),
             " contain multiple elements per row. Please either drop this ",
             "column or reduce its elements to a single value per row.")
    idx <- match(colnames(spd), names(mapping))
    colnames(spd)[!is.na(idx)] <- mapping[idx[!is.na(idx)]]
    if (any(colnames(spd) == "CHARGE")) {
        sign_char <- ifelse(spd$CHARGE > 0, "+", "-")
        nas <- is.na(spd$CHARGE)
        spd$CHARGE <- paste0(abs(spd$CHARGE), sign_char)
        spd$CHARGE[nas] <- ""
    }
    if (!exportTitle && any(colnames(spd) %in% "TITLE"))
        spd$TITLE <- NULL
    l <- nrow(spd)
    tmp <- lapply(colnames(spd), function(z) {
        paste0(z, "=", spd[, z], "\n")
    })
    if (exportTitle && !any(colnames(spd) %in% "TITLE")) {
        sn <- spectraNames(x)
        if (any(sn != as.character(seq_along(x))))
            title <- paste0("TITLE=", sn, "\n")
        else
            title <- paste0("TITLE=msLevel ", spd$msLevel, "; retentionTime ",
                            spd$rtime, "; scanNum ", spd$acquisitionNum, "\n")
        tmp <- c(list(title), tmp)
    }
    pks <- vapply(peaksData(x), function(z)
        paste0(paste0(z[, 1], " ", z[, 2], "\n"), collapse = ""),
        character(1))
    tmp <- do.call(cbind, c(list(rep_len("BEGIN IONS\n", l)),
                            tmp, list(pks), list(rep_len("END IONS\n", l))))
    tmp[grep("=NA\n", tmp)] <- ""
    writeLines(apply(tmp, 1, paste0, collapse = ""), con = con)
}

test_that("readMgf works", {
    fls <- dir(system.file("extdata", package = "MsBackendMgf"),
               full.names = TRUE, pattern = "mgf$")

    expect_error(readMgf(fls), "Please provide a single mgf file.")

    res1 <- readMgf(fls[1])
    cns <- c("rtime", "acquisitionNum", "precursorMz", "precursorIntensity",
             "precursorCharge", "mz", "intensity", "dataOrigin",
             "msLevel", "TITLE")
    expect_identical(sort(names(res1)), sort(cns))
    expect_true(is(res1$mz, "NumericList"))
    expect_true(is(res1$intensity, "NumericList"))
    expect_equal(length(res1$intensity[[1]]), length(res1$mz[[1]]))
    expect_equal(length(res1$intensity[[2]]), length(res1$mz[[2]]))
    expect_equal(length(res1$intensity[[3]]), length(res1$mz[[3]]))
    expect_identical(res1$TITLE,
                     c("File193 Spectrum1719 scans: 2162",
                       "File193 Spectrum1944 scans: 2406",
                       "File193 Spectrum1968 scans: 2432"))

    res2 <- readMgf(fls[2])
    cns <- c("rtime", "precursorMz", "precursorIntensity",
             "precursorCharge", "mz", "intensity", "dataOrigin",
             "msLevel", "TITLE")
    expect_identical(sort(names(res2)), sort(cns))
    expect_true(is(res2$mz, "NumericList"))
    expect_true(is(res2$intensity, "NumericList"))
    expect_equal(length(res2$intensity[[1]]), length(res2$mz[[1]]))
    expect_equal(length(res2$intensity[[2]]), length(res2$mz[[2]]))
    expect_equal(length(res2$intensity[[3]]), length(res2$mz[[3]]))
    expect_true(is.na(res2$rtime[4]))

    res3 <- readMgf(fls[3])
    expect_true(length(res3$mz[[1L]]) == 0)
    expect_true(length(res3$intensity[[1L]]) == 0)
})

test_that("readMgfSplit works", {
    fls <- dir(system.file("extdata", package = "MsBackendMgf"),
               full.names = TRUE, pattern = "mgf$")

    expect_error(readMgfSplit(fls), "Please provide a single mgf file.")
    expect_error(readMgfSplit(fls[1], n = -3), "has to be an integer")
    expect_error(readMgfSplit(fls[1], n = 3), "Consider increasing the value")

    res1 <- readMgfSplit(fls[1])
    cns <- c("rtime", "acquisitionNum", "precursorMz", "precursorIntensity",
             "precursorCharge", "mz", "intensity", "dataOrigin",
             "msLevel", "TITLE")
    expect_identical(sort(names(res1)), sort(cns))
    expect_true(is(res1$mz, "NumericList"))
    expect_true(is(res1$intensity, "NumericList"))
    expect_equal(length(res1$intensity[[1]]), length(res1$mz[[1]]))
    expect_equal(length(res1$intensity[[2]]), length(res1$mz[[2]]))
    expect_equal(length(res1$intensity[[3]]), length(res1$mz[[3]]))
    expect_identical(res1$TITLE,
                     c("File193 Spectrum1719 scans: 2162",
                       "File193 Spectrum1944 scans: 2406",
                       "File193 Spectrum1968 scans: 2432"))
    expect_equal(res1, readMgf(fls[1]))

    res2 <- readMgfSplit(fls[2])
    cns <- c("rtime", "precursorMz", "precursorIntensity",
             "precursorCharge", "mz", "intensity", "dataOrigin",
             "msLevel", "TITLE")
    expect_identical(sort(names(res2)), sort(cns))
    expect_true(is(res2$mz, "NumericList"))
    expect_true(is(res2$intensity, "NumericList"))
    expect_equal(length(res2$intensity[[1]]), length(res2$mz[[1]]))
    expect_equal(length(res2$intensity[[2]]), length(res2$mz[[2]]))
    expect_equal(length(res2$intensity[[3]]), length(res2$mz[[3]]))
    expect_true(is.na(res2$rtime[4]))
    expect_equal(res2, readMgf(fls[2]))

    res3 <- readMgfSplit(fls[3])
    expect_true(length(res3$mz[[1L]]) == 0)
    expect_true(length(res3$intensity[[1L]]) == 0)
    expect_equal(res3, readMgf(fls[3]))
})

test_that(".extract_mgf_spectrum works", {
    fls <- dir(system.file("extdata", package = "MsBackendMgf"),
               full.names = TRUE, pattern = "mgf$")
    mgf <- scan(fls[1], what = "", sep = "\n", quote = "",
                allowEscapes = FALSE, quiet = TRUE)
    cmts <- grep("^[#;!/]", mgf)
    if (length(cmts))
        mgf <- mgf[-cmts]

    begin <- grep("BEGIN IONS", mgf) + 1L
    end <- grep("END IONS", mgf) - 1L
    n <- length(begin)

    res <- .extract_mgf_spectrum(mgf[begin[1]:end[1]])
    expect_true(is.data.frame(res))
    expect_equal(names(res), c("TITLE", "PEPMASS", "CHARGE",
                               "RTINSECONDS", "SCANS", "PEPMASSINT",
                               "mz", "intensity"))
    expect_true(is.na(res$PEPMASSINT))

    with_mocked_bindings(
        ".is_unsorted" = function(...) TRUE,
        code = expect_equal(res, .extract_mgf_spectrum(mgf[begin[1]:end[1]]))
    )

    res <- .extract_mgf_spectrum(mgf[begin[2]:end[2]])
    expect_true(is.data.frame(res))
    expect_equal(names(res), c("TITLE", "PEPMASS", "CHARGE",
                               "RTINSECONDS", "SCANS", "PEPMASSINT",
                               "mz", "intensity"))
    expect_true(!is.na(res$PEPMASSINT))
})

test_that(".format_charge works", {
    res <- .format_charge(c("4+", "3-", NA, "45"))
    expect_equal(res, c("4", "-3", NA_character_, "45"))
})

test_that(".is_unsorted works", {
    x <- cbind(1:3, 1:3)
    expect_false(.is_unsorted(x))

    x <- cbind(c(1, 43, 2), 1:3)
    expect_true(.is_unsorted(x))

    expect_false(.is_unsorted(x[integer(), , drop = FALSE]))
})

test_that(".extract_mgf_spectrum_with_annotations works", {
    fls <- dir(system.file("extdata", package = "MsBackendMgf"),
               full.names = TRUE, pattern = "mgf$")
    a <- grep("spectra.mgf", fls, value = TRUE)
    A <- readLines(a)
    mgf <- A[grep("BEGIN IONS", A)[1L]:grep("END IONS", A)[1L]]
    res <- .extract_mgf_spectrum_with_annotations(mgf)
    expect_true(is.data.frame(res))
    expect_true(length(res$mz) > 0)
    expect_true(length(res$intensity) > 0)
    expect_true(is.data.frame(res$ann_mat[[1L]]))
    expect_true(ncol(res$ann_mat[[1L]]) == 0L)
    expect_equal(nrow(res$ann_mat[[1L]]), length(res$mz[[1L]]))

    b <- grep("fiora", fls, value = TRUE)
    B <- readLines(b)
    mgf <- B[grep("BEGIN IONS", B)[1L]:grep("END IONS", B)[1L]]
    res <- .extract_mgf_spectrum_with_annotations(mgf)
    expect_true(is.data.frame(res))
    expect_true(length(res$mz) > 0)
    expect_true(length(res$intensity) > 0)
    expect_true(length(res$ann_mat) > 0)
    expect_true(is.data.frame(res$ann_mat[[1L]]))
    expect_equal(length(res$intensity[[1L]]), length(res$mz[[1L]]))
    expect_equal(length(res$intensity[[1L]]), nrow(res$ann_mat[[1L]]))

    mgf <- c("BEGIN IONS",
             "TEST=12.3",
             "1 431.1 a b",
             "2 343.1 d",
             "3 313.1",
             "4 542.1 e",
             "END IONS")
    res <- .extract_mgf_spectrum_with_annotations(mgf)
    expect_true(is.data.frame(res))
    expect_true(length(res$mz) > 0)
    expect_true(length(res$intensity) > 0)
    expect_true(length(res$ann_mat) > 0)
    expect_true(is.data.frame(res$ann_mat[[1L]]))
    expect_equal(length(res$intensity[[1L]]), length(res$mz[[1L]]))
    expect_equal(length(res$intensity[[1L]]), nrow(res$ann_mat[[1L]]))
    expect_equal(ncol(res$ann_mat[[1L]]), 2)
})

test_that("readMgf works with annotated = TRUE", {
    fls <- dir(system.file("extdata", package = "MsBackendMgf"),
               full.names = TRUE, pattern = "mgf$")

    b <- grep("fiora", fls, value = TRUE)
    l <- readLines(b)
    l2 <- c(l, l)
    l2[13] <- "888.575459486 0.002360690850764513"
    tf <- tempfile()
    writeLines(l2, tf)
    res <- readMgf(tf, annotated = TRUE)
    expect_s4_class(res, "DataFrame")
    expect_true(nrow(res) == 2)
    expect_equal(res$mz[[1L]], res$mz[[2L]])
    expect_true(is.na(res$V1[[1L]][6]))
    expect_equal(res$V1[[1L]][-6L], res$V1[[2L]][-6L])
    expect_equal(metadata(res), list("V1"))

    ## Manually defining
    mgf <- c("BEGIN IONS",
             "TITLE=a",
             "1 431.1 a b",
             "2 343.1 d",
             "3 313.1",
             "4 542.1 e",
             "END IONS",
             "BEGIN IONS",
             "TITLE=b",
             "1 2 d",
             "2 2",
             "END IONS",
             "BEGIN IONS",
             "TITLE=c",
             "1 3",
             "2 3",
             "3 3",
             "END IONS"
             )
    writeLines(mgf, tf)
    res <- readMgf(tf, annotated = TRUE)
    expect_equal(res$TITLE, c("a", "b", "c"))
    expect_equal(res$V1, list(c("a", "d", NA, "e"),
                              c("d", NA),
                              c(NA_character_, NA_character_, NA_character_)))
    expect_equal(res$V2, list(c("b", NA, NA, NA),
                              c(NA_character_, NA_character_),
                              c(NA_character_, NA_character_, NA_character_)))
    expect_equal(metadata(res), list(c("V1", "V2")))
    file.remove(tf)
})

---
title: "Description and usage of MsBackendMgf"
output:
    BiocStyle::html_document:
        toc_float: true
vignette: >
    %\VignetteIndexEntry{Description and usage of MsBackendMgf}
    %\VignetteEngine{knitr::rmarkdown}
    %\VignetteEncoding{UTF-8}
    %\VignettePackage{MsBackendMgf}
    %\VignetteDepends{Spectra,BiocStyle,BiocParallel}
---

```{r style, echo = FALSE, results = 'asis', message=FALSE}
BiocStyle::markdown()
```

**Package**: `r Biocpkg("MsBackendMgf")`<br />
**Authors**: `r packageDescription("MsBackendMgf")[["Author"]] `<br />
**Last modified:** `r file.info("MsBackendMgf.Rmd")$mtime`<br />
**Compiled**: `r date()`

```{r, echo = FALSE, message = FALSE}
library(Spectra)
library(BiocStyle)
```

# Introduction

The `r Biocpkg("Spectra")` package provides a central infrastructure for the
handling of Mass Spectrometry (MS) data. The package supports interchangeable
use of different *backends* to import MS data from a variety of sources (such as
mzML files). The `r Biocpkg("MsBackendMgf")` package allows the import of MS/MS
data from MGF ([Mascot Generic
Format](http://www.matrixscience.com/help/data_file_help.html)) files. The
`MsBackendMgf` backend allows to load and represent data from these *standard*
MGF files, while the `MsBackendAnnotatedMgf` backend supports import of data
from MGF files containing also annotations for the individual mass peaks. This
vignette illustrates the usage of the *MsBackendMgf* package.

# Installation

To install this package, start `R` and enter:

```{r, eval = FALSE}
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("MsBackendMgf")
```

This will install this package and all eventually missing dependencies.


# Importing MS/MS data from MGF files

MGF files store one to multiple spectra, typically centroided and of MS
level 2. In our short example below, we load MGF files which are provided with
this package. In a next section we also import data from an MGF file containing
individual peak annotations. Below we first load all required packages and
define the paths to the MGF files.

```{r load-libs}
library(Spectra)
library(MsBackendMgf)

fls <- dir(system.file("extdata", package = "MsBackendMgf"),
           full.names = TRUE, pattern = "^spectra(.*).mgf$")
fls
```

MS data can be accessed and analyzed through `Spectra` objects. Below
we create a `Spectra` with the data from these MGF files. To this end
we provide the file names and specify to use a `MsBackendMgf()`
backend as *source* to enable data import.

```{r import}
sps <- Spectra(fls, source = MsBackendMgf())
```

With that we have now full access to all imported spectra variables
that we list below.

```{r spectravars}
spectraVariables(sps)
```

Besides default spectra variables, such as `msLevel`, `rtime`,
`precursorMz`, we also have additional spectra variables such as the
`TITLE` of each spectrum in the MGF file.

```{r instrument}
sps$rtime
sps$TITLE
```

By default, fields in the MGF file are mapped to spectra variable names using
the mapping returned by the `spectraVariableMapping()` function:

```{r spectravariables}
spectraVariableMapping(MsBackendMgf())
```

The names of this `character` vector are the spectra variable names (such as
`"rtime"`) and the field in the MGF file that contains that information are the
values (such as `"RTINSECONDS"`). Note that it is also possible to overwrite
this mapping (e.g. for certain MGF *dialects*) or to add additional
mappings. Below we add the mapping of the MGF field `"TITLE"` to a spectra
variable called `"spectrumName"`.

```{r map}
map <- c(spectrumName = "TITLE", spectraVariableMapping(MsBackendMgf()))
map
```

We can then pass this mapping to the `backendInitialize()` method, or the
`Spectra()` constructor.

```{r import2}
sps <- Spectra(fls, source = MsBackendMgf(), mapping = map)
```

We can now access the spectrum's title with the newly created spectra variable
`"spectrumName"`:

```{r spectrumName}
sps$spectrumName
```

In addition we can also access the m/z and intensity values of each
spectrum.

```{r mz}
mz(sps)
intensity(sps)
```

The `MsBackendMgf` backend allows also to export data in MGF format. Below we
export the data to a temporary file. We hence call the `export()` function on
our `Spectra` object specifying `backend = MsBackendMgf()` to use this backend
for the export of the data. Note that we use again our custom mapping of
variables such that the spectra variable `"spectrumName"` will be exported as
the spectrums' title.

```{r export}
fl <- tempfile()
export(sps, backend = MsBackendMgf(), file = fl, mapping = map)
```

We next read the first lines from the exported file to verify that the title was
exported properly.

```{r export-check}
readLines(fl)[1:12]
```

Note that the `MsBackendMgf` exports all spectra variables as fields in the mgf
file. To illustrate this we add below a new spectra variable to the object and
export the data.

```{r}
sps$new_variable <- "A"
export(sps, backend = MsBackendMgf(), file = fl)
readLines(fl)[1:12]
```

We can see that also our newly defined variable was exported. Also, because we
did not provide our custom variable mapping this time, the variable
`"spectrumName"` was **not** used as the spectrum's title.

Sometimes it might be required to not export all spectra variables since some
exported fields might not be recognized/supported by external tools. Using the
`selectSpectraVariables()` function we can reduce our `Spectra` object to export
to contain only relevant spectra variables. Below we restrict the data to only
m/z, intensity, retention time, acquisition number, precursor m/z and precursor
charge and export these to an MGF file. Also, some external tools don't support
the `"TITLE"` field in the MGF file. To disable export of the spectrum ID/title
`exportTitle = FALSE` can be used.

```{r}
sps_ex <- selectSpectraVariables(sps, c("mz", "intensity", "rtime",
                                        "acquisitionNum", "precursorMz",
                                        "precursorCharge"))
export(sps_ex, backend = MsBackendMgf(), file = fl, exportTitle = FALSE)
readLines(fl)[1:12]
```

## Annotated MGF files

Variants of MGF files can also contain annotations for the individual mass
peaks. These files are expected to contain one line per mass peak, with the
first two elements being the mass peak's *m/z* and intensity values. All
(eventually present) additional elements are considered *annotations*. The
`MsBackendAnnotatedMgf` backend can be used to import and represent such files.
Below we read the first 10 lines of an example annotated MGF file.

```{r}
fl <- dir(system.file("extdata", package = "MsBackendMgf"),
          pattern = "fiora", full.names = TRUE)
readLines(fl, 10)
```

The last 3 displayed lines show the mass peak content of a MGF spectrum with the
*m/z*, intensity and an annotation for the mass peak. Such data files can be
imported using the `MsBackendAnntotatedMgf` backend:

```{r}
sps_ann <- Spectra(fl, source = MsBackendAnnotatedMgf())
sps_ann
```

The peaks annotation have also been imported and are available as additional
*peaks variables*:

```{r}
peaksVariables(sps_ann)
```

The names of the annotations follow the standard R convention of naming columns
in a `data.frame` if no name is provided, i.e. it consists of `"V"` followed by
a number. The data can be retrieved with the `peaksData()` function specifying
the names of the peaks variables to extract. If not provided, `peakData()` will
only extract the *m/z* and intensity values. Thus, below we call `peaksData()`
providing the names of all requested variables (columns).

```{r}
pd <- peaksData(sps_ann, columns = c("mz", "intensity", "V1"))
```

The peaks data of one spectrum is now represented as a `data.frame`:

```{r}
pd[[1L]] |>
    head()
```

# Parallel processing

The *MsBackendMgf* package supports parallel processing for data
import. Parallel processing can be enabled by providing the parallel processing
setup to the `backendInitialize()` function (or the `readMgf()` function) with
the `BPPARAM` parameter. By default (with `BPPARAM = SerialParam()`) parallel
processing is disabled. If enabled, and data import is performed on a single
file, the extraction of spectra information on the imported MGF file is
performed in parallel. If data is to be imported from multiple files, the import
is performed in parallel on a per-file basis (i.e. in parallel from the
different files). Generally, the performance gain through parallel processing is
only moderate and it is only suggested if a large number of files need to be
processed, or if the MGF file is very large (e.g. containing over 100,000
spectra).

# Session information

```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hidden_aliases.R
\docType{methods}
\name{hidden_aliases}
\alias{hidden_aliases}
\alias{[,MsBackendDataFrame-method}
\title{Internal page for hidden aliases}
\description{
For S4 methods that require a documentation entry but only clutter the index.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MsBackendMgf.R
\name{MsBackendMgf}
\alias{MsBackendMgf}
\alias{MsBackendMgf-class}
\alias{backendInitialize,MsBackendMgf-method}
\alias{spectraVariableMapping,MsBackendMgf-method}
\alias{export,MsBackendMgf-method}
\alias{backendInitialize,MsBackendAnnotatedMgf-method}
\alias{MsBackendAnnotatedMgf}
\title{MS data backend for MGF files}
\usage{
\S4method{backendInitialize}{MsBackendMgf}(
  object,
  files,
  mapping = spectraVariableMapping(object),
  nlines = -1L,
  ...,
  BPPARAM = SerialParam()
)

MsBackendMgf()

\S4method{spectraVariableMapping}{MsBackendMgf}(object, format = c("mgf"))

\S4method{export}{MsBackendMgf}(
  object,
  x,
  file = tempfile(),
  mapping = spectraVariableMapping(object),
  exportTitle = TRUE,
  ...
)

\S4method{backendInitialize}{MsBackendAnnotatedMgf}(
  object,
  files,
  mapping = spectraVariableMapping(object),
  ...,
  BPPARAM = SerialParam()
)

MsBackendAnnotatedMgf()
}
\arguments{
\item{object}{Instance of \code{MsBackendMgf} class.}

\item{files}{\code{character} with the (full) file name(s) of the mgf file(s)
from which MS/MS data should be imported.}

\item{mapping}{for \code{backendInitialize()} and \code{export}: named \code{character}
vector allowing to specify how fields from the MGF file should be
renamed. Names are supposed to be the spectra variable name and
values of the vector the field names in the MGF file. See output of
\code{spectraVariableMapping()} for the expected format and examples
below or description above for details.}

\item{nlines}{for \code{backendInitialize()} of \code{MsBackendMgf}: \code{integer(1)}
defining the number of lines that should be imported and processed
from the MGF file(s).
By default (\code{nlines = -1L}) the full file is imported and processed at
once. If set to a positive integer, the data is imported and processed
\emph{chunk-wise} using \code{\link[=readMgfSplit]{readMgfSplit()}}.}

\item{...}{Currently ignored.}

\item{BPPARAM}{Parameter object defining the parallel processing
setup. If parallel processing is enabled (with \code{BPPARAM} different than
\code{SerialParam()}, the default) and length of \code{files} is larger than one,
import is performed in parallel on a per-file basis. If data is to be
imported from a single file (i.e., length of \code{files} is one), parsing
of the imported file is performed in parallel. See also
\code{\link[BiocParallel:SerialParam-class]{BiocParallel::SerialParam()}} for information on available parallel
processing setup options.}

\item{format}{for \code{spectraVariableMapping()}: \code{character(1)} defining the
format to be used. Currently only \code{format = "mgf"} is supported.}

\item{x}{for \code{export()}: an instance of \code{\link[Spectra:Spectra]{Spectra::Spectra()}} class with the
data that should be exported.}

\item{file}{\code{character(1)} with the (full) file name to which the data
should be exported.}

\item{exportTitle}{\code{logical(1)} whether the \emph{TITLE} field should be included
in the exported MGF file. If \code{TRUE} (the default) a \code{spectraVariable}
called \code{"TITLE"} will be used, if no such variable is present either the
\code{spectraNames(object)} will be used or, if they are empty, a title will
be generated including the MS level, retention time and acquisition
number of the spectrum.}
}
\value{
See description above.
}
\description{
The \code{MsBackendMgf} class supports import and export of MS/MS spectra data
from/to files in Mascot Generic Format
(\href{http://www.matrixscience.com/help/data_file_help.html}{mgf})
files. After initial import, the full MS data is kept in
memory. \code{MsBackendMgf} extends the \code{\link[Spectra:MsBackend]{Spectra::MsBackendDataFrame()}} backend
directly and supports thus the \code{\link[Spectra:addProcessing]{Spectra::applyProcessing()}} function to make
data manipulations persistent.

The \code{MsBackendAnnotatedMgf} class supports import of data from MGF files
that provide, in addition to the \emph{m/z} and intensity values, also
additional annotations/metadata for each mass peak. For such MGF files it
is expected that each line contains information from a single mass peak,
separated by a white space (blank). The first two elements are expected to
be the peak's \emph{m/z} and intensity values, while each additional element is
considered an annotation for this specific peak. See examples below for the
format of a supported MGF file. The \code{backendInitialize()} method of
\code{MsBackendAnnotatedMgf} does not support parameter \code{nlines}. Also, import
of data can be considerably slower compared to the standard \code{MsBackendMgf}
backend, because of the additionally required parsing of peak annotations.
Peaks information in MGF files are not named, thus, additional peaks
annotations are named using the standard naming convention for column named
of data frames: the first peaks annotation is called \code{"V1"}, the second (if
available) \code{"V2"} and so on.

New objects are created with the \code{MsBackendMgf()} or
\code{MsBackendAnnotatedMgf()} function. The \code{backendInitialize()} method has to
be subsequently called to initialize the object and import the MS/MS data
from (one or more) MGF files.

The \code{MsBackendMgf} backend provides an \code{export} method that allows to export
the data from the \code{Spectra} object (parameter \code{x}) to a file in mgf format.
See the package vignette for details and examples.

Default mappings from fields in the MGF file to spectra variable names are
provided by the \code{spectraVariableMapping()} function. This function returns a
named character vector were names are the spectra variable names and the
values the respective field names in the MGF files. This named character
vector is submitted to the import and export function with parameter
\code{mapping}. It is also possible to pass own mappings (e.g. for special
MGF dialects) with the \code{mapping} parameter.
}
\examples{

library(BiocParallel)
#' Getting the file names of all example MGF files from MsBackendMgf
fls <- dir(system.file("extdata", package = "MsBackendMgf"),
    full.names = TRUE, pattern = "^spectra(.*).mgf$")

## Create an MsBackendMgf backend and import data from test mgf files.
be <- backendInitialize(MsBackendMgf(), fls)
be

be$msLevel
be$intensity
be$mz

## The spectra variables that are available; note that not all of them
## have been imported from the MGF files.
spectraVariables(be)

## The variable "TITLE" represents the title of the spectrum defined in the
## MGF file
be$TITLE

## The default mapping of MGF fields to spectra variables is provided by
## the spectraVariableMapping function
spectraVariableMapping(MsBackendMgf())

## We can provide our own mapping e.g. to map the MGF field "TITLE" to a
## variable named "spectrumName":
map <- c(spectrumName = "TITLE", spectraVariableMapping(MsBackendMgf()))
map

## We can then pass this mapping with parameter `mapping` to the
## backendInitialize method:
be <- backendInitialize(MsBackendMgf(), fls, mapping = map)

## The title is now available as variable named spectrumName
be$spectrumName

## Next we create a Spectra object with this data
sps <- Spectra(be)

## We can use the 'MsBackendMgf' also to export spectra data in mgf format.
out_file <- tempfile()
export(sps, backend = MsBackendMgf(), file = out_file, map = map)

## The first 20 lines of the generated file:
readLines(out_file, n = 20)

## Next we add a new spectra variable to each spectrum
sps$spectrum_idx <- seq_along(sps)

## This new spectra variable will also be exported to the mgf file:
export(sps, backend = MsBackendMgf(), file = out_file, map = map)
readLines(out_file, n = 20)

####
## Annotated MGF

## An example of a supported annotated MGF file
fl <- system.file("extdata", "xfiora.mgf", package = "MsBackendMgf")

## Lines with peak data start with a numeric and information is
## separated by a whitespace. The first two elements are the peak's m/z
## and intensity while any additional information is considered as
## annotation. Information for each peak is provided in one line.
readLines(fl)

## Importing the data using an `MsBackendAnnotatedMgf`
ba <- backendInitialize(MsBackendAnnotatedMgf(), fl)
ba

## An additional peaks variable is available.
peaksVariables(ba)

ba$V1

## The length of such peaks variables is the same as the length of the
## m/z or intensity values, i.e. each peak has one value (with the value
## being `NA` if missing).
length(ba$V1[[1L]])
length(ba$mz[[1L]])

## Extracting the peaks data from a `Spectra` with a `MsBackendAnnotatedMgf`
s <- Spectra(ba)
pd <- peaksData(s, peaksVariables(ba))[[1L]]
head(pd)
class(pd)
}
\author{
Laurent Gatto, Corey Broeckling and Johannes Rainer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions-mgf.R
\name{readMgf}
\alias{readMgf}
\alias{readMgfSplit}
\title{Reading MGF files}
\usage{
readMgf(
  f,
  msLevel = 2L,
  mapping = spectraVariableMapping(MsBackendMgf()),
  annotated = FALSE,
  ...,
  BPPARAM = SerialParam()
)

readMgfSplit(
  f,
  msLevel = 2L,
  mapping = spectraVariableMapping(MsBackendMgf()),
  nlines = 1e+05,
  BPPARAM = SerialParam(),
  ...
)
}
\arguments{
\item{f}{\code{character(1)} with the path to an mgf file.}

\item{msLevel}{\code{numeric(1)} with the MS level. Default is 2.}

\item{mapping}{named \code{character} vector to rename mgf fields to spectra
variables.}

\item{annotated}{For \code{readMgf()}: \code{logical(1)} whether the MGF file
contains additional peak annotations. See examples below or the
documentation for \code{\link[=MsBackendAnnotatedMgf]{MsBackendAnnotatedMgf()}} for information on the
expected format.}

\item{...}{Additional parameters, currently ignored.}

\item{BPPARAM}{parallel processing setup that should be used. Only the
parsing of the imported MGF file is performed in parallel.}

\item{nlines}{for \code{readMgfSplit()}: \code{integer(1)} with the number of lines
that should be imported and parsed in each iteration.}
}
\value{
A \code{DataFrame} with each row containing the data from one spectrum
in the MGF file. m/z and intensity values are available in columns \code{"mz"}
and \code{"intensity"} in a list representation. For \code{readMgf()} with
\code{annotated = TRUE} also all peaks annotation columns (named \verb{"V1", etc) are provided in a list representation, with the lengths of elements matching those of }"mz"\code{or}"intensity"`.
}
\description{
The \code{readMgf()} function imports the data from a file in MGF format reading
all specified fields and returning the data as a \code{\link[S4Vectors:DataFrame-class]{S4Vectors::DataFrame()}}.

For very large MGF files the \code{readMgfSplit()} function might be used
instead. In contrast to the \code{readMgf()} functions, \code{readMgfSplit()} reads
only \code{nlines} lines from an MGF file at once reducing thus the memory
demand (at the cost of a lower performance, compared to \code{readMgf()}).
}
\examples{

fls <- dir(system.file("extdata", package = "MsBackendMgf"),
    full.names = TRUE, pattern = "mgf$")[1L]

readMgf(fls)

## Annotated MGF
fl <- system.file("extdata", "xfiora.mgf", package = "MsBackendMgf")
res <- readMgf(fl, annotated = TRUE)
colnames(res)
res$V1
}
\author{
Laurent Gatto, Johannes Rainer, Sebastian Gibb, Corey Broeckling
}
